using System;
using System.Collections;
using System.Drawing;
using System.Reflection;
using System.IO;

using org.zefer.pd4ml;
using org.zefer.html.parser;
using org.zefer.html.doc;
using org.zefer.pd4ml.pdf;

namespace PD4Test
{
    class Tokenizer
    {
        //Element list identified
        private System.Collections.ArrayList elements;
        //Source string to use
        private string source;
        //The tokenizer uses the default delimiter set: the space character, the tab character, the newline character, and the carriage-return character
        private string delimiters = " \t\n\r";      

        /// <summary>
        /// Initializes a new class instance with a specified string to process
        /// </summary>
        /// <param name="source">String to tokenize</param>
        public Tokenizer(string source)
        {           
            this.elements = new System.Collections.ArrayList();
            this.elements.AddRange(source.Split(this.delimiters.ToCharArray()));
            this.RemoveEmptyStrings();
            this.source = source;
        }

        /// <summary>
        /// Initializes a new class instance with a specified string to process
        /// and the specified token delimiters to use
        /// </summary>
        /// <param name="source">String to tokenize</param>
        /// <param name="delimiters">String containing the delimiters</param>
        public Tokenizer(string source, string delimiters)
        {
            this.elements = new System.Collections.ArrayList();
            this.delimiters = delimiters;
            this.elements.AddRange(source.Split(this.delimiters.ToCharArray()));
            this.RemoveEmptyStrings();
            this.source = source;
        }

        /// <summary>
        /// Current token count for the source string
        /// </summary>
        public int Count
        {
            get
            {
                return (this.elements.Count);
            }
        }

        /// <summary>
        /// Determines if there are more tokens to return from the source string
        /// </summary>
        /// <returns>True or false, depending if there are more tokens</returns>
        public bool HasMoreTokens()
        {
            return (this.elements.Count > 0);           
        }

        /// <summary>
        /// Returns the next token from the token list
        /// </summary>
        /// <returns>The string value of the token</returns>
        public string NextToken()
        {           
            string result;
            if (source == "") throw new System.Exception();
            else
            {
                this.elements = new System.Collections.ArrayList();
                this.elements.AddRange(this.source.Split(delimiters.ToCharArray()));
                RemoveEmptyStrings();       
                result = (string) this.elements[0];
                this.elements.RemoveAt(0);              
                this.source = this.source.Remove(this.source.IndexOf(result),result.Length);
                this.source = this.source.TrimStart(this.delimiters.ToCharArray());
                return result;                  
            }           
        }

        /// <summary>
        /// Returns the next token from the source string, using the provided
        /// token delimiters
        /// </summary>
        /// <param name="delimiters">String containing the delimiters to use</param>
        /// <returns>The string value of the token</returns>
        public string NextToken(string delimiters)
        {
            this.delimiters = delimiters;
            return NextToken();
        }

        /// <summary>
        /// Removes all empty strings from the token list
        /// </summary>
        private void RemoveEmptyStrings()
        {
            for (int index=0; index < this.elements.Count; index++)
                if ((string)this.elements[index]== "")
                {
                    this.elements.RemoveAt(index);
                    index--;
                }
        }
    }

    /// <summary>
    /// Summary description for Class1.
    /// </summary>
    class PD4Test
    {
        const string USAGE = "Usage: pd4net.exe '<url>' <htmlWidth> [pageFormat] [-permissions <NUMBER>]" + 
            "[-bookmarks <HEADINGS|ANCHORS>] [-orientation <PORTRAIT|LANDSCAPE>] " + 
			"[-insets <T,L,B,R,><mm|pt>] [-ttf <ttf_fonts_dir>] [-out <output_file_path>] " + 
			"[-type PDF|RTF|Image]";

		private static void wrongParam(string msg) 
		{
            Console.Out.WriteLine("invalid parameter: " + msg);
            Console.Out.WriteLine( USAGE );
        }

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main(string[] args)
        {
            //System.Threading.Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-US");
            //System.Threading.Thread.CurrentThread.CurrentUICulture = System.Threading.Thread.CurrentThread.CurrentCulture;
			try 
			{
				Main_Impl(args);
			}
			catch(Exception e) 
			{
				Console.Error.WriteLine(e.Message + ":" + e.StackTrace);
			}
		}

		static void Main_Impl(string[] args)
		{
            String infile = null;
            int width = -1;
            String format = "A4";
            int permissions = -1;
            String bookmarks = null;
            String orientation = "portrait";
            String insets = null;
            String ttf = null;
            String outfile = null;
			String type = "PDF";
			bool debugOn = false;

            bool formatSet = false;
            FileInfo ff;
            
			for( int i = 0; i < args.Length; i++ ) 
			{
				if ( "-permissions".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing permissions number (a sum of single permission codes)");
                        return; 
                    }
                    permissions = Int32.Parse(args[i]);
                    continue;
                }

				if ( "-type".Equals( args[i] ) ) 
				{
					++i;
					if ( i == args.Length ) 
					{
						wrongParam("missing type, one of PDF RTF Image");
						return; 
					}
					type = args[i];
					continue;
				}

				if ( "-bookmarks".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing bookmark type (HEADINGS or ANCHORS)");
                        return; 
                    }
                    bookmarks = args[i];
                    continue;
                }

				if ( "-orientation".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing orientation type (PORTRAIT or LANDSCAPE)");
                        return; 
                    }
                    orientation = args[i];
                    continue;
                }

				if ( "-insets".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing insets (for example: -insets 5,10,5,5mm)");
                        return; 
                    }
                    insets = args[i];
                    continue;
                }

				if ( "-ttf".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing TTF fonts directory");
                        return; 
                    }
                    ff = new FileInfo(args[i]);
                    ttf = ff.FullName;
                    continue;
                }

				if ( "-out".Equals( args[i] ) ) 
				{
                    ++i;
					if ( i == args.Length ) 
					{
                        wrongParam("missing output file name");
                        return; 
                    }
                    outfile = args[i];
                    continue;
                }

				if ( args[i].Equals("-debug")  ) 
				{
					debugOn = true;
					continue;
				}

				if ( args[i].StartsWith("-")  ) 
				{
                    wrongParam("unknown: " + args[i]);
                    return; 
                }

				if ( infile == null ) 
				{
                    string ss = args[i];
					if(!ss.StartsWith("http://") && !ss.StartsWith("file://")) 
					{
                        ff = new FileInfo(ss);
                        infile = "file:///" + ff.FullName;
                    }
                    else
                        infile = ss;
                    continue;
                }

				if ( width == -1 ) 
				{
                    width = Int32.Parse(args[i]);
                    continue;
                }

				if ( !formatSet ) 
				{
                    formatSet = true;
                    format = args[i];
                    continue;
                }

                wrongParam(args[i]);
                return; 
            }
            
			if ( infile == null ) 
			{
                Console.Out.WriteLine("source URL is missing");
                Console.Out.WriteLine( USAGE );
                return;
            }

			if ( width == -1 ) 
			{
                Console.Out.WriteLine("HTML width parameter is missing");
                Console.Out.WriteLine( USAGE );
                return;
            }

            PD4Test converter = new PD4Test();

			if(type == null || type.Length == 0)
				type = "PDF";
			type = type.Trim().ToUpper();
			
			if(type.Equals("PDF"))
                converter.generatePDF( 
                    infile, width, format, permissions, bookmarks, orientation, 
					insets, ttf, outfile, debugOn ); 
			else if(type.Equals("RTF"))
				converter.generateRTF(infile, width, insets, outfile, debugOn);
			else if(type.Equals("IMAGE"))
				converter.generatePNG(infile, format, insets, outfile, debugOn);
			else
				wrongParam("wrong type: " + type);
        }
        
		private bool coverInsets(String insets, ref Rectangle rr, ref char uom)
        {
			if(insets == null) return false;
                Tokenizer st = new Tokenizer( insets, ",");
			rr = new Rectangle(0,0,0,0);
			uom = 'P';
			try 
			{
                int top = Int32.Parse(st.NextToken());
                int left = Int32.Parse(st.NextToken());
                int bottom = Int32.Parse(st.NextToken());
				String rightStr = st.NextToken();
				String units = null;
				int right = 0;
                    
				if(st.Count > 4) 
				{
					right = Int32.Parse(rightStr);
					units = st.NextToken();
				}
				else 
				{ 
					units = rightStr.Substring(rightStr.Length - 2, 2);
					rightStr = rightStr.Substring(0, rightStr.Length -2);
					right = Int32.Parse(rightStr);
				}

				rr.X = left;
				rr.Y = top;
				rr.Width = right - left;
				rr.Height = bottom - top;
				if ("MM".Equals(units.ToUpper())) 
				{
					uom = 'M';
				} 
				else 
				{
					uom = 'P';
                }
			} 
			catch (Exception) 
			{
				rr = new Rectangle(0,0,0,0);
				uom = 'P';
                    throw new Exception(
                            "Invalid page insets (top, left, bottom, right, units): "
                                    + insets );
                }
			return true;
		}
        
		private void generatePDF(String inputUrl, int htmlWidth, String pageFormat, int permissions, 
			String bookmarks, String orientation, String insets, String fontsDir, String outfile, bool debugOn)
		{
			PD4ML pd4ml = new PD4ML();

			Rectangle ins = new Rectangle(0,0,0,0);
			char uom = 'P';
			if(coverInsets(insets, ref ins, ref uom)) 
			{
				if (uom == 'M') 
				{
					pd4ml.PageInsetsMM = ins;
				} 
				else 
				{
					pd4ml.PageInsets = ins;
				}
            }
            
            pd4ml.HtmlWidth = (htmlWidth);
            if(orientation == null) orientation = "portrait";
            if ("PORTRAIT".Equals(orientation.ToUpper())) 
                pd4ml.PageSize = PD4Constants.getSizeByName(pageFormat);
            else 
                pd4ml.PageSize = pd4ml.changePageOrientation( PD4Constants.getSizeByName(pageFormat) ); 
            
            if ( permissions != -1 ) pd4ml.setPermissions("empty", permissions, true);
			if ( bookmarks != null ) 
			{
                if ( "ANCHORS".Equals(bookmarks.ToUpper()) )
                    pd4ml.generateOutlines(false);
                else if ( "HEADINGS".Equals(bookmarks.ToUpper()) ) 
                    pd4ml.generateOutlines(true);
            }

            if ( fontsDir != null && fontsDir.Length > 0 ) pd4ml.useTTF( fontsDir, true );
			if (outfile == null || outfile.Equals("-"))
			{ 
				pd4ml.render(new Uri(inputUrl), Console.Out);
			}
			else
				pd4ml.render(
					new Uri(inputUrl), 
					new System.IO.FileStream(outfile, System.IO.FileMode.Create));
        }

		private void generateRTF(String inputUrl, int htmlWidth, String insets, String outfile, bool debugOn) 
		{
			PD4MLHtmlParser parser = new PD4MLHtmlParser(new Uri(inputUrl), null, null, null, null, 
				false, null, null, debugOn, null, null, -1, false);
			Rectangle ins = new Rectangle(20,40,-20,0);
			char uom = 'P';
			if(coverInsets(insets, ref ins, ref uom) && uom == 'M') ins = new Rectangle(20,40,-20,0);

			Document doc = parser.buildDocument();
			doc.layout(htmlWidth);
			int height = doc.Height;
        
			PD4Device dev = new PD4Device(htmlWidth, PD4Constants.A4, ins);
			dev.DevType = PD4Device.DevType_RTF;
			doc.paint(0, 0, new Rectangle(0, 0, htmlWidth, height), dev);
			dev.MyDispose();

			if (outfile == null || outfile.Equals("-"))
				Console.Out.WriteLine(dev.ToString());
			else 
			{
				FileStream fs = new FileStream(outfile, FileMode.Create);
				byte[] bb = SupportClass.getBytes(dev.ToString());
				fs.Write(bb, 0, bb.Length);
				fs.Close();
			}
		}
        
		private void generatePNG(String inputUrl, String format, String insets, String outfile, bool debugOn) 
		{
			Rectangle ins = new Rectangle(20,40,-20,0);
			char uom = 'P';
			if(coverInsets(insets, ref ins, ref uom) && uom == 'M') ins = new Rectangle(20,40,-20,0);

			Size pageS = PD4Constants.getSizeByName(format);
			PD4MLHtmlParser parser = new PD4MLHtmlParser(new Uri(inputUrl), null, null, null, null, 
				false, null, null, debugOn, null, null, -1, false);

			Document doc = parser.buildDocument();
			doc.layout(pageS.Width);
			int documentHeight = doc.Height;
        
			PD4Device dev = new PD4Device(pageS.Width, pageS.Height, documentHeight, 0);
			dev.DevType = PD4Device.DevType_Image;
			int pc = 0;
			int deviceScroll = 0;

			do 
			{
				if ( pc > 0 ) 
				{
					dev.startNewPage();
					deviceScroll = dev.CurrentOffset;
				}

				dev.BgColor = Color.White;
				dev.FillRectangle(dev.Brush, 0, 0, pageS.Width, pageS.Height);
				doc.paint(0, -deviceScroll, new Rectangle(0, -deviceScroll, pageS.Width, pageS.Height + deviceScroll), dev);
				pc++;
			} while ( dev.NewPageNeeded );
			ArrayList bitmaps = dev.Result;

			if (outfile == null || outfile.Equals("-")) 
			{
				if(bitmaps.Count > 0)
				{ 
					MemoryStream ms = new MemoryStream();
					encodeImage((Bitmap) bitmaps[0], ms);
					ms.Close();
					byte[] b = ms.ToArray();
					Console.Out.Write(SupportClass.newString(b, 0, b.Length, Console.Out.Encoding));
				}
			}
			else 
			{
				int j = outfile.LastIndexOf('.');
				if(j < 0) 
				{
					outfile += ".png";
					j = outfile.LastIndexOf('.');
				}

				String fnbase = outfile.Substring(0, j);
				String fnext = outfile.Substring(j);
				for(j=0;j<bitmaps.Count; ++j) 
				{
					Bitmap bmp = (Bitmap)bitmaps[j];
					FileStream fs = new FileStream(fnbase + "-" + j + fnext, FileMode.Create);
					encodeImage(bmp, fs);
					fs.Close();
				}
			}
		}

		private void encodeImage(Bitmap bmp, Stream fs)
		{
			bmp.Save(fs, System.Drawing.Imaging.ImageFormat.Png);
		}
    }
}
